package FUNSEQ;

use 5.000;
use strict;
use List::Util qw[min max];
use List::Util qw(sum);

our $VERSION = '0.1';

=head1 NAME

	FUNSEQ - Function based prioritization of sequence variants 

=head1 SYNOPSIS

	use FUNSEQ;
	FUNSEQ::function

=head1 METHOD

        format_check()
                Input format check.
                
        snv_filter ()
                Filter variants using minor allele frequency threshold based on 1000 genomes phase1. 
                
        motif_break()
                Check whether the variants break motifs. 
     
        sensitive()
                Locate in sensitive/ultra-sensitive regions or not. 
                
        gene_link()
                Link non-coding variants with genes (promoter -2.5kb & distal regulatory module).
  
  		coding()
                Coding Pipe - with VAT (variant annotation tool) pipe
                
        intergrate()
                Interage coding & noncoding outputs, and generate ordered output
               
        recur()
                Recurrent calculation
                Arguments : Input Genomes; Output Path; Output File
                
=head1 AUTHOR
        
    Yao Fu ; Gerstein Lab; Yale University

=head1 COPYRIGHT AND LICENSE

	Copyright (C) 2013 by  Gerstein Lab

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.1 or,
at your option, any later version of Perl 5 you may have available.

=cut



## Input format checking 
sub format_check{
# Arguments taken : input file, input format ;

	my ($input,$format) = @_;
	my $err_bed = <<EOF;

Error: incorrect bed format; please prepare your file as :
chromosome	start	end	reference_allele	alternative_allele	...
chr1  213941196  213942363	G	T	...
chr1  213942363  213943530	A	C	...
chr1  213943530  213944697	T	A	...
EOF

	my $err_vcf = <<EOF;

Error: incorrect vcf format; please prepare your file as :
#CHROM POS    ID        REF  ALT     QUAL FILTER INFO ...
2      4370   rs6057    G    A       29   .      ...   ...
EOF

	my $status = 0;
	my @lines;
	my $i;
	
	if ($format =~ /bed/i){
		@lines = split /\n/,`head -n 1000 $input`;
		for  $i(0 .. $#lines){
			if ($lines[$i] =~ /^chr\S+\s+\d+\s+\d+\s+\S+\s+\S+/i){
			}else{
				print $lines[$i],"\n";
				$status = 1;
			}
		}
		if ($status == 1){
			print $err_bed;
		}
		return $status;
	}
	if ($format =~ /vcf/i){
		@lines = split /\n/,`head -n 1000 $input|grep -v '#'`;
		for $i (0 .. $#lines){
			if ($lines[$i] =~ /\S+\s+\d+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+/i){
			}else{
				print $lines[$i],"\n";
				$status=1;
			}
		}
		if ($status == 1){
			print $err_vcf;
		}
		return $status;
	}
}

## Constructor 
sub new {
    my $class=shift;
    my $self ={};
    bless($self,$class);
    return $self;
}
    
## Filter variants against 1000 genomes phase 1 with minor allele frequency...
sub snv_filter{
# Arguments taken : input file ,input format ,1000 genome file, minor allele frequency input ,output
	
	my $self = shift;
	my @ARGV = @_;
	unless (scalar @ARGV == 5){
	print <<EOF;
	This script is to filter SNVs in 1000 Phase1 project.
	snv_filter : Incorrect number of arguments ... 
EOF
	exit;	
	}
	
	my $input = $ARGV[0];   #Input File
	my $input_format = $ARGV[1];  #Input File Format
	my $tgp_snp = $ARGV[2];    #1KG snp file
	my $maf_cut_off  = $ARGV[3];  #MAF cut-off
	my $output = $ARGV[4];   #output
	
	
	my $ErrorHead = "ERROR: snv_filter"; 
	my $input_line; 
	my @snp_tabix;
	my $snp;
	my $id;
	my %saw;
	
	open(OUT,">$output")||die;
	`sed -i 's/^chr//gI' $input`;
	`sed -i -e '/^#/! s/^/chr/' $input`;

	@snp_tabix = split /\n/, `tabix $tgp_snp -B $input | awk '{FS="\t";OFS="\t"} \$4 >= $maf_cut_off'`;
	foreach $snp (@snp_tabix){
		$id = join("\t",(split /\s+/,$snp)[0..1]);
		$id =~ s/chr//;
		$saw{$id} =1;
	}
	if ($input_format =~ /vcf/i){
		open(IN,$input)|| die "Input file not found ...\n";
		while(<IN>){
		   	if (!/^#/){
		   		my @tmp = split /\s+/,$_;
		   		# only SNVs
		   		if (length($tmp[3])==1 && length($tmp[4])==1 && $tmp[3]=~ /[ATCG]/i && $tmp[4]=~ /[ATCG]/i){
		   			$tmp[0]=~ s/chr//;
		   			my $id = join("\t",$tmp[0],$tmp[1]-1);
		   			if(defined $saw{$id}){
		   			}else{
		   				$self->{DES} ->{$id} = join("\t",@tmp[0 ..7]);
		   				print OUT "chr",$tmp[0],"\t",join("\t",@tmp[1 ..6]),"\t.\n";
		   			}
		   		}
		   	}
		}
		close IN;
	}elsif($input_format =~ /bed/i){
		open(IN,$input)|| die "Input file not found ...\n";
		while(<IN>){
		   	my @tmp = split /\s+/,$_;
		   	# only SNVs
		   	if (length($tmp[3])==1 && length($tmp[4])==1 && $tmp[3]=~ /[ATCG]/i && $tmp[4]=~ /[ATCG]/i){
		   		my $id = join("\t",$tmp[0],$tmp[1]);
		   		$id =~ s/chr//g;
		   		if(defined $saw{$id}){
		   		}else{
		   			$self->{DES} -> {$id} = join("\t",@tmp[0 ..4]);
		   			print OUT join("\t",@tmp[0 ..4]),"\n";
		   		}
		   	}
		}
		close IN;
	}else{
		die "$ErrorHead : unknown format, please use vcf or bed format";	
	}
	close OUT;
}

## Motif Analysis
sub motif_break{
# Arguments taken : input_file, input_format, ancestral_file, encode_motif_data, somatic/germline_mode, motif_pfm
	my $self = shift;
	
	local @ARGV=@_;
	unless (scalar @ARGV==6){
	print <<EOF;
		motif_break: Missing argument...
		out_annotated_nc, informat, ancestral_path, bound_motif, genome_mode, motif_pfm, out_motif_break
EOF
	exit;
	}

	$| = 1;
	
	my $input_file = $ARGV[0];
	my $input_format = $ARGV[1];
	my $ancestral_file = $ARGV[2];
	my $bound_motif = $ARGV[3];
	my $mode = $ARGV[4];
	my $pfm = $ARGV[5];

	my $input_file_new;
	my $id;
	my %ancestral;
	my $chr;
	my $pos;
	my $line;
	my @info;
	my $temp;
	my $prev_name;
	my %motif;
	my $A = 1;
	my $C = 2;
	my $G = 3;
	my $T = 4;
	my $ref;
	my $alt;
	my $AA;
	my $der_al;
	my $motif_len;
	my $factor;
	my $pos_in_motif;
	my $motif_name;
	my $derived_allele_freq;
	my $AA_freq;
	my $interval;

# Read in PFM file 
	open MOTIF, $pfm or die;
	while(<MOTIF>){
		chomp $_;
		if(/^>/){
			$prev_name = (split/>|\s+/,$_)[1];
		}else{
			@info = split/\s+/,$_;
			if(not exists $motif{$prev_name}){
				$motif{$prev_name}->[0] = {(A=>$info[$A], T=>$info[$T], C=>$info[$C], G=>$info[$G])};
			}else{
				$temp = $motif{$prev_name};
				$motif{$prev_name}->[scalar(@$temp)] = {(A=>$info[$A], T=>$info[$T], C=>$info[$C], G=>$info[$G])};			
			}
		}
	}
	close MOTIF;


	if ($input_format =~ /bed/i){
		$input_file_new = `cut -f 1,2,3,4,5 $input_file | intersectBed -a stdin -b $bound_motif -wo | sort -k 1,1 -k 2,2n |uniq`;
		if ($mode == 2){
			my @line = split /\n/, `printf "$input_file_new" | sed "s,^chr,," |awk '{FS="\t";OFS="\t"}{print \$1,\$2,\$3}' | fastaFromBed -fi $ancestral_file -bed stdin -fo stdout`;
			foreach my $line (@line){
				if ($line =~ />/){
					($chr, $pos) = (split />|:|-/,$line)[1..2];
					$id = join("",$chr,"\t",$pos);
				}else{
					chomp $line;
					$ancestral{$id} =$line;
				}
			}
		}
	
		foreach my $input_line (split /\n/,$input_file_new){
			chomp $input_line;
			@info = split /\t+/,$input_line;
			$ref = uc($info[3]); $alt = uc($info[4]);
			$id = join("\t",$info[0],$info[1]);
			$id =~ s/chr//g;

		
			if($mode == 1){
				$AA = $ref;
				$der_al = $alt;
			}else{
				$AA = uc($ancestral{$id});			
				if($AA eq $ref){
					$der_al = $alt;
				}elsif($AA eq $alt){
					$der_al = $ref;
				}else{
					next;
				}
			}
		
			$motif_len = $info[7]-$info[6];
			$factor = $info[11];
			if($info[10] eq "+"){
				$pos_in_motif = ($info[1]-$info[6]+1);
			}elsif($info[10] eq "-"){
				$pos_in_motif = $info[7]-$info[1];
				$AA =~ tr/ACGTacgt/TGCAtgca/;
				$der_al =~ tr/ACGTacgt/TGCAtgca/;
			}
		
			$motif_name = (split/_\d+mer/,$info[8])[0];
			if(not exists $motif{$motif_name}){
				print "Motif_Name_Not_Found\t$line\n";
				next;
			}
			my $ref = $motif{$motif_name};
			if($motif_len != scalar(@$ref)){
				print "Motif_Len_Not_Matched\t$line\n";
				next;
			}
		
			$derived_allele_freq = $motif{$motif_name}->[$pos_in_motif-1]->{$der_al};
			$AA_freq = $motif{$motif_name}->[$pos_in_motif-1]->{$AA};
			$interval = join("",$info[5],":",$info[6],"-",$info[7]);
			$self -> {ANNO} ->{$id} -> {"TPM($factor|$interval)"}=1;
			if ($derived_allele_freq < $AA_freq){
				if (defined $self -> {MOTIFBR}->{$id}){
					$self -> {MOTIFBR} -> {$id} = join(",",$self->{MOTIFBR}->{$id}, "$factor#$info[8]#$info[6]#$info[7]#$info[10]#$pos_in_motif#$derived_allele_freq#$AA_freq");
				}else{
					$self->{MOTIFBR}->{$id} = "$factor#$info[8]#$info[6]#$info[7]#$info[10]#$pos_in_motif#$derived_allele_freq#$AA_freq"; 
				}
			}		
		}
		close IN;
	
	}elsif($input_format =~ /vcf/i){
		$input_file_new = `grep -v '#' $input_file | cut -f 1,2,3,4,5,6,7,8 | intersectBed -a stdin -b $bound_motif -wo | sort -k 1,1 -k 2,2n |uniq`;
		if ($mode == 2){
			my @line = split /\n/, `printf "$input_file_new" | sed "s,^chr,," |awk '{FS="\t";OFS="\t"}{print \$1,\$2-1,\$2}' | fastaFromBed -fi $ancestral_file -bed stdin -fo stdout`;
			foreach my $line (@line){
				if ($line =~ />/){
					($chr, $pos) = (split />|:|-/,$line)[1 ..2];
					$id = join("",$chr,"\t",$pos);
				}else{
					chomp $line;
					$ancestral{$id} =$line;
				}
			}
		}
		
		foreach my $input_line (split /\n/,$input_file_new){
			chomp $input_line;
			@info = split /\t+/,$input_line;
			$ref = uc($info[3]); $alt = uc($info[4]);
			$id = join("\t",$info[0],$info[1]-1);
			$id =~ s/chr//g;


			if($mode == 1){
				$AA = $ref;
				$der_al = $alt;
			}else{
				$AA = uc($ancestral{$id});			
				if($AA eq $ref){
					$der_al = $alt;
				}elsif($AA eq $alt){
					$der_al = $ref;
				}else{
					next;
				}
			}
		
			$motif_len = $info[10]-$info[9];
			$factor = $info[14];
			if($info[13] eq "+"){
				$pos_in_motif = $info[1]-$info[9];
			}elsif($info[13] eq "-"){
				$pos_in_motif = $info[10]-$info[1]+1;
				$AA =~ tr/ACGTacgt/TGCAtgca/;
				$der_al =~ tr/ACGTacgt/TGCAtgca/;
			}
		
			$motif_name = (split/_\d+mer/,$info[11])[0];
			if(not exists $motif{$motif_name}){
				print "Motif_Name_Not_Found\t$line\n";
				next;
			}
			my $ref = $motif{$motif_name};
			if($motif_len != scalar(@$ref)){
				print "Motif_Len_Not_Matched\t$line\n";
				next;
			}
		
			$derived_allele_freq = $motif{$motif_name}->[$pos_in_motif-1]->{$der_al};
			$AA_freq = $motif{$motif_name}->[$pos_in_motif-1]->{$AA};
			$interval = join("",$info[8],":",$info[9],"-",$info[10]);
			$self -> {ANNO} ->{$id} ->{"TPM($factor|$interval)"}=1;
			if ($derived_allele_freq < $AA_freq){
				if(defined $self ->{MOTIFBR}->{$id}){
					$self -> {MOTIFBR} -> {$id} =join(",",$self->{MOTIFBR}->{$id}, "$factor#$info[11]#$info[9]#$info[10]#$info[13]#$pos_in_motif#$derived_allele_freq#$AA_freq");
				}else{
					$self -> {MOTIFBR}->{$id} = "$factor#$info[11]#$info[6]#$info[7]#$info[10]#$pos_in_motif#$derived_allele_freq#$AA_freq";
				}
			}	
		}
		close IN;
	}
}

## Sensitive
sub sensitive{
	my $self =shift;
	my ($input,$sensitive,$format) = @_;
	my @sen;
	my @ultra;
	my $id;
	
	@sen = split /\n/, `intersectBed -u -a  $input -b $sensitive`;
	@ultra = split /\n/, `grep "Ultra" $sensitive| intersectBed -u -a  $input -b stdin`;
	
	if ($format =~ /bed/i){
		foreach my $line (@sen){
			$id = join("\t",(split /\t+/,$line)[0..1]);
			$id =~ s/chr//g;
			$self->{SEN}->{$id} =1;			
		}	
		foreach my $line (@ultra){
			$id = join("\t",(split /\t+/,$line)[0..1]);
			$id =~ s/chr//;
			$self -> {USEN}->{$id}=1;
		}
	}else{
		foreach my $line(@sen){
			my @tmp = split /\s+/,$line;
			$id = join("\t",$tmp[0],$tmp[1]-1);
			$id =~ s/chr//;
			$self->{SEN}->{$id}=1;
		} 
		foreach my $line(@ultra){
			my @tmp = split /\s+/,$line;
			$id = join("\t",$tmp[0],$tmp[1]-1);
			$id =~ s/chr//;
			$self->{USEN}->{$id}=1;
		}
	}
}

## Promoter/Enhancer 
sub gene_link{
	my $self = shift;
	my ($input,$format,$promoter,$enhancer) = @_;
	my @pro;
	my @enh;
	my $line;

	@pro =split /\n/, `intersectBed -a $input -b $promoter -wo | sort -k 1,1 -k 2,2n |uniq`;
	@enh = split /\n/, `intersectBed -a $input -b $enhancer -wo | sort -k 1,1 -k 2,2n |uniq`;
	
	if ($format =~ /bed/i){
		foreach $line(@pro){
			my @tmp = split /\s+/,$line;
			my $id = join("\t",$tmp[0],$tmp[1]);
			my $gene = $tmp[8];
			$id =~ s/chr//;
			$self->{GENE}->{$id} -> {"$gene(Promoter)"} =1;
			if ($tmp[9] =~ /PPI/ && $tmp[9] =~ /REG/){
				$self->{HUB} -> {$id} -> {"PPI&REG($gene)"}=1;
			}elsif($tmp[9] =~ /PPI/){
				$self->{HUB} -> {$id} -> {"PPI($gene)"}=1;
			}elsif($tmp[9] =~ /REG/){
				$self->{HUB} -> {$id} -> {"REG($gene)"}=1;
			}
		}
		foreach $line(@enh){
			my @tmp = split /\s+/,$line;
			my $id = join("\t",$tmp[0],$tmp[1]);
			$id =~ s/chr//;
			my $gene_tag = $tmp[8];
			my ($gene, $hub_info) = (split /\(/,$gene_tag)[0..1];
			$self->{GENE}->{$id} -> {"$gene(DRM)"} =1;
			if ($hub_info =~ /PPI/ && $hub_info =~ /REG/){
				$self->{HUB} -> {$id} -> {"PPI&REG($gene)"}=1;
			}elsif($hub_info =~ /PPI/){
				$self->{HUB} -> {$id} -> {"PPI($gene)"}=1;
			}elsif($hub_info =~ /REG/){
				$self->{HUB} -> {$id} -> {"REG($gene)"}=1;
			}
		}
	}elsif($format =~ /vcf/i){
		foreach $line(@pro){
			my @tmp = split /\s+/,$line;
			my $id = join("\t",$tmp[0],$tmp[1]-1);
			my $gene = $tmp[11];
			$id =~ s/chr//;
			$self->{GENE}->{$id} -> {"$gene(Promoter)"} =1;
			if ($tmp[12] =~ /PPI/ && $tmp[12] =~ /REG/){
				$self->{HUB} -> {$id} -> {"PPI&REG($gene)"}=1;
			}elsif($tmp[12] =~ /PPI/){
				$self->{HUB} -> {$id} -> {"PPI($gene)"}=1;
			}elsif($tmp[12] =~ /REG/){
				$self->{HUB} -> {$id} -> {"REG($gene)"}=1;
			}
		}
		foreach $line(@enh){
			my @tmp = split /\s+/,$line;
			my $id = join("\t",$tmp[0],$tmp[1]-1);
			my $gene_tag = $tmp[11];
			my ($gene,$hub_info) = (split /\(/,$gene_tag)[0..1];
			$id =~ s/chr//;
			$self->{GENE}->{$id} -> {"$gene(DRM)"} =1;
			if ($hub_info =~ /PPI/ && $hub_info =~ /REG/){
				$self->{HUB} -> {$id} -> {"PPI&REG($gene)"}=1;
			}elsif($hub_info =~ /PPI/){
				$self->{HUB} -> {$id} -> {"PPI($gene)"}=1;
			}elsif($hub_info =~ /REG/){
				$self->{HUB} -> {$id} -> {"REG($gene)"}=1;
			}
		}
	}
}

## Coding : vat analysis
sub coding{
	local @ARGV= @_;
	unless(scalar @ARGV == 10){
	print <<EOF;
		funseq.coding.pl: Missing argument ...
EOF
	exit;
}

	my $snp_input = $ARGV[0];
	my $format = $ARGV[1];
	my $file_interval = $ARGV[2];
	my $file_fasta = $ARGV[3];
	my $file_ppi = $ARGV[4];
	my $file_reg = $ARGV[5];
	my $file_selection = $ARGV[6];
	my $out = $ARGV[7];
	my $out_path = $ARGV[8];
	my $nc_mode=$ARGV[9];
	my @tmp;
	my %ppi_hub;
	my %reg_hub;
	my %selection;
	my $gene;
	my @vat_out;
	my $vat_out;
	
	if ($nc_mode ==0){
# run VAT
	if ($format =~ /vcf/i){
		@vat_out = split /\n/, `awk '{FS="\t";OFS="\t"}BEGIN{print "##fileformat=VCFv4.0\\n#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO"}{print \$1,\$2,".",\$4,\$5,".","PASS","."}' $snp_input| snpMapper $file_interval $file_fasta`;
	}

	if ($format =~ /bed/i){
		@vat_out = split /\n/, `awk '{FS="\t";OFS="\t"}BEGIN{print "##fileformat=VCFv4.0\\n#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO"}{print \$1,\$3,".",\$4,\$5,".","PASS","."}' $snp_input| snpMapper $file_interval $file_fasta`;
	}


#### incorporate gene selection / LoF and Hubs ### 
	open(IN,$file_ppi)||die;
	while(<IN>){
		$gene = (split /\s+/,$_)[0];
		$ppi_hub{$gene}=1;	
	}
	close IN;

	open(IN,$file_reg)||die;
	while(<IN>){
		$gene = (split /\s+/,$_)[0];
		$reg_hub{$gene} =1;
	}
	close IN;


	open(IN,$file_selection)||die;
	while(<IN>){
		$gene = (split /\s+/,$_)[0];
		$selection{$gene}=1;
	}
	close IN;


	if(scalar @vat_out ==0){
		return "1";
	}else{
		open(O,">$out")||die;
		foreach $vat_out(@vat_out){
			if ($vat_out =~ /^#/){
			}else{
				@tmp =split /\s+/,$vat_out;
				print O $tmp[0],"\t",$tmp[1],"\t",$tmp[7],"\t";
				my $info = (split /,/,$tmp[7])[0];
				if ($info =~ /VA=\d+:(\S+):ENSG/){
					$gene =$1;
					print O $gene,"\t";
					if (defined $ppi_hub{$gene}){
						print O "PPI;";
					}else{
						print O ";";
					}
					if (defined $reg_hub{$gene}){
						print O "REG;";
					}else{
						print O ";";
					}
					if (defined $selection{$gene}){
						print O "NegativeSelection\n";
					}else{
						print O "\n";
					}
				}
			}
		}
	close O;
	return "0";
	}
	}else{
		return "0";
	}
}

## Integration of all outputs
sub intergrate{
	my $self=shift;
	local @ARGV= @_;
	unless (scalar @ARGV == 7){
print <<EOF;
	funseq.intergrate.pl : Missing arguments ...
EOF
exit;		
}

	$| = 1; #Auto Flush

	my $input_format = $ARGV[0];  # Input Format
	my $output_format = $ARGV[1]; # Output Format
	my $tag = $ARGV[2]; # tag 
	my $nc_snp = $ARGV[3];  # noncoding_snp 
	my $encode = $ARGV[4];
	my $coding_info = $ARGV[5];
	my $out = $ARGV[6];

	open(OUT,">$out")||die;
	my %nc_score;
	my %cds_score;
	my %cancer;
	my %cancer_reg;

	my $vcf_header = <<EOF;
##fileformat=VCFv4.0
##INFO=<ID=OTHER,Number=.,Type=String, Description = "Other Information From Original File">
##INFO=<ID=CDS,Number=.,Type=String,Description="Coding Variants or not">
##INFO=<ID=VA,Number=.,Type=String,Description="Coding Variant Annotation">
##INFO=<ID=HUB,Number=.,Type=String,Description="Network Hubs, PPI (protein protein interaction network), REG (regulatory network)">
##INFO=<ID=GNEG,Number=.,Type=String,Description="Gene Under Negative Selection">
##INFO=<ID=NCENC,Number=.,Type=String,Description="NonCoding ENCODE Annotation">
##INFO=<ID=MOTIFBR,Number=.,Type=String,Description="Motif Breaking">
##INFO=<ID=SEN,Number=.,Type=String,Description="In Sensitive Region">
##INFO=<ID=USEN,Number=.,Type=String,Description="In Ultra-Sensitive Region">
##INFO=<ID=GENE,Number=.,Type=String,Description="Target Gene (For coding - directly affected genes ; For non-coding - promoter or enhancer)">
##INFO=<ID=CANG,Number=.,Type=String,Description="Known Cancer gene / TF regulating Known Cancer gene";
##INFO=<ID=CDSS,Number=.,Type=String,Description="FunSEQ Coding Score">
##INFO=<ID=NCDS,Number=.,Type=String,Description="FunSEQ NonCoding Score">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO
EOF


	my $bed_header=join("\t","chr","start","end","ref","alt","cds;variant.annotation.cds;network.hub;gene.under.negative.selection;ENCODE.annotated;motif.breaking;sensitive;ultra.sensitive;target.gene;coding.score;noncoding.score");

	&read_nc($nc_snp,$input_format);
	&read_encode($encode,$input_format);
	&read_cds($coding_info);
	&nc_score_scheme();
	&rank_output($output_format);
 
 
 	sub read_nc{
		my ($file,$format) = @_;
		if ($format =~ /bed/i ){
			open(IN,$file);
			while(<IN>){	
				chomp $_;
				my @tmp = split /\s+/,$_;
				my $id = join("\t",$tmp[0],$tmp[1]);
				$id =~ s/chr//;
				$self ->{NC} ->{$id} =1;
			}
			close IN;
		}elsif($format =~ /vcf/i){
			open (IN, $file);
			while(<IN>){
				chomp $_;
				my @tmp = split /\s+/,$_;
				my $id = join("\t",$tmp[0],$tmp[1]-1);
				$id =~ s/chr//;
				$self->{NC} -> {$id} =1;
			}
			close IN;
		}
	}
   

	sub read_encode{
		my ($file,$format)= @_;
		open(IN,$file);
		while(<IN>){
			chomp $_;
			my @tmp = split /\t+/,$_;
			my $anno;
			my $id;
			my $interval;
			if ($format =~ /bed/i){
				$anno = $tmp[8];
				$id = join("\t",$tmp[0],$tmp[1]);
				$interval = join("",$tmp[5],":",$tmp[6],"-",$tmp[7]);
			}else{
				$anno = $tmp[11];
				$id = join("\t",$tmp[0],$tmp[1]-1);
				$interval = join("",$tmp[8],":",$tmp[9],"-",$tmp[10]);
			}
			$id =~ s/chr//;
			my $tag = "";
			if ($anno =~ /^MCV/){
				$tag = "DHS($anno|$interval)";
			}elsif ($anno =~ /^TFP/){
				my $peak = (split /\./,$anno)[1];
				$tag = "TFP($peak|$interval)";
			}elsif ($anno =~ /^ENST/){
				$tag = "Pseudogene($anno)";
			}elsif($anno=~ /^CHMM/){
				$tag = "Enhancer(chmm/segway|$interval)";
			}elsif($anno=~ /^DRM/){
				$tag = "Enhancer(drm|$interval)";
			}else{
				$tag = "ncRNA($anno)";
			}
			$self->{ANNO}->{$id} -> {$tag}=1;
		}
		close IN;
	}


	sub read_cds{
		my ($file) = @_;
		open(IN,$file);
		while(<IN>){
			chomp $_;
			my @tmp = split /\t+/,$_;
			my $id = join("\t",$tmp[0],$tmp[1]-1);
			$id =~ s/chr//;
			my @info_field = split /;/,$tmp[2];
			my $vat = $info_field[$#info_field];
			my $score = 0;
			$self -> {VAT} ->{$id}= $vat;
			$self ->{GENE} -> {$id} = $tmp[3];
			if (/nonsynonymous/ || /prematureStop/){ ## cannot detect splice mutations, add in next version ... it belongs to noncoding
				if (/prematureStop/){
					$score ++;
				}
				if (/PPI/ && /REG/){
					$self -> {HUB} ->{$id} = "PPI,REG";
					$score ++;
				}elsif (/PPI/){
					$self -> {HUB} ->{$id} = "PPI";
					$score ++;
				}elsif(/REG/){
					$self -> {HUB} ->{$id} = "REG";
					$score ++;
				}
				if (/NegativeSelection/){
					$self -> {SELECTION} ->{$id} = 1;
					$score ++;
				}	
				if (defined $self ->{UCONS} ->{$id}){
    				$score++;
    			}
    			
				$cds_score{$score}{$id} = 1;
			}elsif(/removeStop/ || /splice/){
				$self -> {OTHER} ->{$id} = 1;
			}else{
				$self -> {OTHER} -> {$id}=1;
			}
		}
		close IN;
	}

	sub rank_output{
		# rank coding infront of noncoding 
		my ($format) = @_;
		if ($format =~ /bed/i){
    		print OUT $bed_header,"\n";
    		&print_bed_format();
    	}elsif($format =~ /vcf/i){
			print OUT $vcf_header;
			&print_vcf_format;
   		}else{
    		return "Please specify proper output format";	
    	}
	}

	sub print_bed_format{
		 #### Coding #####
		my $score;
		my $id;
    	foreach $score (sort {$b<=>$a} keys %cds_score){
    		foreach $id (sort keys %{$cds_score{$score}}){
    			if ($input_format =~ /bed/i){
    				print OUT $self -> {DES} -> {$id},"\t";
    			}else{
    				my @tmp = split /\s+/,$self ->{DES} ->{$id};
    				print OUT $tmp[0],"\t",$tmp[1]-1,"\t",$tmp[1],"\t",$tmp[3],"\t",$tmp[4],"\t";
    			}
    			print OUT "Yes;";
    			
    			print OUT $self -> {VAT} ->{$id},";";
    			if (defined $self ->{HUB} ->{$id}){
    				print OUT $self ->{HUB} ->{$id},";";
    			}else{
    				print OUT ".;";
    			}
    			if (defined $self ->{SELECTION} ->{$id}){
    				print OUT "Yes;";
    			}else{
    				print OUT ".;";
    			}
    			print OUT ".;.;.;.;";
    			print OUT $self -> {GENE} -> {$id},";";
    			print OUT $score+1,";";
    			print OUT ".\n";
    		}
    	}
    
    	foreach $id(sort keys %{$self -> {OTHER}}){
    		if ($input_format =~ /bed/i){
    			print OUT $self -> {DES} -> {$id},"\t";
    		}else{
    			my @tmp = split /\s+/,$self -> {DES} -> {$id};
    			print OUT $tmp[0],"\t",$tmp[1]-1,"\t",$tmp[1],"\t",$tmp[3],"\t",$tmp[4],"\t";
    		}
    		print OUT "Yes",";";
    		print OUT $self -> {VAT} ->{$id},";";
    		print OUT ".;"x6;
    		print OUT $self -> {GENE} -> {$id},";";
    		print OUT "0;.\n";
    	}
    
   
    ######## Non-coding ############
    
    	foreach $score (sort {$b<=>$a} keys %nc_score){
    		foreach $id (sort keys %{$nc_score{$score}}){
    			if ($input_format =~ /bed/i){
    				print OUT $self -> {DES} ->{$id},"\t";
    			}else{
    				my @tmp = split /\s+/,$self -> {DES} ->{$id};
    				print OUT $tmp[0],"\t",$tmp[1]-1,"\t",$tmp[1],"\t",$tmp[3],"\t",$tmp[4],"\t";
    			}
    			print OUT "No",";";
    			print OUT ".;";
    			if (defined $self -> {HUB} -> {$id}){
    				print OUT join(",",sort keys %{$self -> {HUB} -> {$id}}),";";
    			}else{
    				print OUT ".;";
    			}
    			print OUT ".;";   # negative selection
    		
    			if (defined $self -> {ANNO} -> {$id}){
    				print OUT join(",",sort keys %{$self -> {ANNO} -> {$id}}),";";
    			}else{
    				print OUT ".;";
    			}
    			if (defined $self -> {MOTIFBR} -> {$id}){
    				print OUT $self -> {MOTIFBR} -> {$id},";";
    			}else{
    				print OUT ".;";
    			}
    		
    			if (defined $self -> {SEN} -> {$id}){
    				print OUT "Yes;";
    			}else{
    				print OUT ".;";
    			}
    			if (defined $self -> {USEN} -> {$id}){
    				print OUT "Yes;";
    			}else{
    				print OUT ".;";
    			}
    			
    			
    			if (defined $self -> {GENE} -> {$id}){
    				print OUT join(",",keys %{$self -> {GENE} -> {$id}}),";";
    			}else{
    				print OUT ".;";
    			}
    			print OUT ".;";
    			print OUT $score,"\n";
    		}
    	}
	}

	sub print_vcf_format{
	#### Coding ####
		my $score;
		my $id;
    	foreach $score (sort {$b<=>$a} keys %cds_score){
    		foreach $id (sort keys %{$cds_score{$score}}){
    			if ($input_format =~ /vcf/i){
    				print OUT $self->{DES}->{$id},";";
    			}else{
    				my @tmp = split /\t+/,$self->{DES}->{$id};
    				print OUT $tmp[0],"\t",$tmp[2],"\t.\t",$tmp[3],"\t",$tmp[4],"\t",".\t.\t";
    			}
    			print OUT "CDS=Yes",";";
    			print OUT $self -> {VAT} -> {$id},";";
    			if (defined $self -> {HUB}->{$id}){
    				print OUT "HUB=",$self -> {HUB}->{$id},";";
    			}
    			if (defined $self -> {SELECTION} -> {$id}){
    				print OUT "GNEG=Yes;";
    			}
    			print OUT "GENE=",$self -> {GENE} -> {$id},";";
    			print OUT "CDSS=",$score+1,"\n";
    		}
    	}
    
    	foreach $id(sort keys %{$self -> {OTHER}}){
 			if ($input_format =~ /vcf/i){
    			print OUT $self -> {DES}->{$id},";";
    		}else{
    			my @tmp = split /\t+/,$self->{DES} -> {$id};
    			print OUT $tmp[0],"\t",$tmp[2],"\t.\t",$tmp[3],"\t",$tmp[4],"\t",".\t.\t";
    		}
    		print OUT "CDS=Yes;";
    		print OUT $self -> {VAT} ->{$id},";";
    		print OUT "GENE=",$self -> {GENE} -> {$id},";";
    		print OUT "CDSS=0\n";
    	}
    
   
    ####### Non-coding ###
    
    	foreach $score (sort {$b<=>$a} keys %nc_score){
    		foreach $id (sort keys %{$nc_score{$score}}){
    			if ($input_format =~ /vcf/i){
    				print OUT $self -> {DES} -> {$id},";";
    			}else{
    				my @tmp = split /\t+/,$self -> {DES} ->{$id};
    				print OUT $tmp[0],"\t",$tmp[2],"\t.\t",$tmp[3],"\t",$tmp[4],"\t",".\t.\t";
    			}
    			print OUT "CDS=No;";
    			if (defined $self -> {HUB} -> {$id}){
    				print OUT "HUB=",join(",", sort keys %{$self -> {HUB} -> {$id}}),";";
    			}
    			if (defined $self -> {ANNO} -> {$id}){
    				print OUT "NCENC=",join(",", sort keys %{$self -> {ANNO}->{$id}}),";";
    			}
    			if (defined $self ->{MOTIFBR} ->{$id}){
    				print OUT "MOTIFBR=",$self ->{MOTIFBR}->{$id},";";
    			}
    			if (defined $self ->{SEN}->{$id}){
    				print OUT "SEN=Yes;";
    			}
    			if (defined $self ->{USEN}->{$id}){
    				print OUT "USEN=Yes;";
    			}
    			if (defined $self ->{GENE}->{$id}){
    				print OUT "GENE=", join(",", sort keys %{$self->{GENE}->{$id}}),";";
    			}
    			print OUT "NCDS=",$score,"\n";
    		}
    	}
	}


	sub nc_score_scheme{
		my $score;
		my $id;
		foreach $id (keys %{$self -> {NC}}){
			$score = 0;
			if (defined $self->{ANNO}->{$id}){
				$score ++;
			}
			if (defined $self -> {SEN} -> {$id}){
				$score ++;
			}
			if (defined $self -> {USEN} ->{$id}){
				$score ++;
			}
			if (defined $self -> {GENE} -> {$id}){
				$score ++;
			}
			if (defined $self ->{HUB} ->{$id}){
				$score++;
			}
			if (defined $self -> {MOTIFBR} -> {$id}){
				$score++;
			}
			$nc_score{$score}{$id} = 1;
		}
	}
}

## Recurrent Analysis
sub recur{
	my ($file,$path,$outformat) = @_;
	my @files = @{$file};  #
    my $tag;  #
    my @tag;    #
    my $input;   #
    my @temp;   #
    my $anno;  #
    my $id;  #
    my $elm_info;   #
	my $id_info;  # store id and genome info  
	 
	
	foreach $file(@files){
		@temp = split /\//,$file;
		$tag = (split /\./,$temp[$#temp])[0];
		push @tag,$tag;
		$input = "$path/$tag.FunSEQ.$outformat";

		if ($outformat =~ /bed/i){
			open(IN, $input)||die;
			my $n =0;
			while(<IN>){
				if ($n == 0){
					$n = 1;
				}else{
					@temp = split /\s+/,$_;
					$id = join(":",$temp[0],$temp[2]);
					$id_info->{$id}->{$tag} =1;
					my @feature = split /;/,$temp[5];
					if ($feature[0] eq "Yes" && (/nonsynonymous/ || /prematureStop/)){
						$elm_info -> {$feature[8]}->{$tag}->{$id}=1;
					}elsif($feature[0] eq "No"){
						foreach $anno((split /,/,$feature[4])){
							if ($anno=~ /MCV-/ || $anno eq '.'){
							}else{
								$elm_info -> {$anno} -> {$tag} -> {$id}=1;
							}
						}
					}
				}	
			}
		}else{
			open(IN, $input)||die;
			while(<IN>){
				if (/^#/){
				}else{
					@temp = split /\s+/,$_;
					$id = join(":",$temp[0],$temp[1]);
					$id_info->{$id}->{$tag} =1;
					if (/CDS=Yes/ && (/nonsynonymous/ || /prematureStop/)){
						if (/GENE=(.*?);/){
							$anno = $1;
							$elm_info -> {$anno} -> {$tag} -> {$id}= 1;
						}
					}elsif(/CDS=No/ && /NCENC=/){
						if (/NCENC=(.*?);/){
							foreach $anno((split /,/,$1)){
								if ($anno =~ /MCV-/){
								}else{
									$elm_info -> {$anno} -> {$tag} -> {$id}=1;
								}
							}
						}
					}
				}
			}
		}
	}
	
	
  ## Produce recurrent output summary; 
  	my $join_output;
  	if (scalar @tag <=2){
		$join_output = join("", join("_",@tag),".FunSEQ.recur.Summary");
	}else{
		$join_output = join("", join("_",@tag[0..1]),".FunSEQ.recur.Summary");
	}
	
	open(JOIN,">$path/$join_output")||die;
	my $recur_score;
	my %recur_elm;
	my %recur_id_anno;
	my %recur_id_noanno;
	my $score;
	
	foreach $anno (sort keys %{$elm_info}){
		$score = scalar keys %{$elm_info -> {$anno}};
		if ($score > 1){
			$recur_score -> {$score} -> {$anno} =1;
		}
	}
		
	foreach $score (sort {$b <=> $a} keys %{$recur_score}){
		foreach $anno (sort keys %{$recur_score -> {$score}}){
			print JOIN $anno,"\t";
			my $line = "";
			foreach $tag(sort keys %{$elm_info -> {$anno}}){
				$line = $line."$tag(";
				foreach $id (sort keys %{$elm_info->{$anno}->{$tag}}){
					if (scalar keys %{$id_info->{$id}} > 1){
						$line = $line."$id*,";
						$recur_id_anno{$id} =1;
					}else{
						$line = $line."$id,";
					}
				}
				$line =~ s/,$//;
				$line = $line."),";
			}
			$line =~ s/,$//;
			print JOIN $line,"\n";
			$recur_elm{$anno} = join("",$anno,":",$line);
		}
	}
	
	foreach $id (keys %{$id_info}){
		if (scalar keys %{$id_info->{$id}} >1 && defined $recur_id_anno{$id} != 1){
			print JOIN "None\t$id*(",join(",",keys %{$id_info->{$id}}),")\n";
			$recur_id_noanno{$id} = join ("", "$id*(",join(",",keys %{$id_info->{$id}}),")");
		}
	}
	close JOIN;
	
	#Done 


	
	foreach $tag (@tag){
		$input = "$path/$tag.FunSEQ.$outformat";
		my $output = "$path/$tag.FunSEQ.recur.$outformat";
		if (scalar keys %recur_elm ==0 && scalar keys %recur_id_noanno ==0){
			`mv $input $output`;
		}else{
			open(OUT,">$output")||die;
			if ($outformat =~ /bed/i){
			open(IN, $input)||die;
				my $n =0;
				while(<IN>){
					if ($n ==0 ){
						chomp $_;
						print OUT $_,";noncoding.recurrent\n";
						$n = 1;
					}else{
						chomp $_;
						@temp = split /\s+/,$_;
						$id = join(":",$temp[0],$temp[2]);
						my @feature = split (/;/,$temp[5]);
						if ($feature[0] eq "Yes"){
							$anno = $feature[8];
							if (defined $recur_elm{$anno}){
								print OUT join("\t",@temp[0..4]),"\t",join(";",@feature[0..8]),";",$feature[9]+1,";",$feature[10],";",$recur_elm{$anno},"\n";
							}else{
								print OUT $_,";.\n";
							}
						}else{
							if ($feature[4] eq '.'){
								if(defined $recur_id_noanno{$id}){	
									print OUT join("\t",@temp[0..4]),"\t",join(";",@feature[0..8]),";",$feature[9],";",$feature[10]+1,";",$recur_id_noanno{$id},"\n";
								}else{
									print OUT $_,";.\n";
								}
							}else{
								my $line = "";
								foreach $anno((split /,/,$feature[4])){
									if (defined $recur_elm{$anno}){
										$line = join("",$line,$recur_elm{$anno},",");
									}
								}
								if (length($line )>0){
									$line =~ s/,$//;
									print OUT join("\t",@temp[0..4]),"\t",join(";",@feature[0..8]),";",$feature[9],";",$feature[10]+1,";",$line,"\n";
								}elsif(defined $recur_id_noanno{$id}){	
									print OUT join("\t",@temp[0..4]),"\t",join(";",@feature[0..8]),";",$feature[9],";",$feature[10]+1,";",$recur_id_noanno{$id},"\n";
								}else{
									print OUT $_,";.\n";
								}
							}
						}
					}
				}
			}else{
				open(IN,$input)||die;
				while(<IN>){
					if (/^##/){
						print OUT $_;
					}elsif (/^#CHROM/){
						print OUT '##INFO=<ID=RECUR,Number=.,Type=String,Description="Recurrent elements / variants">',"\n";
						print OUT $_;
					}else{
						@temp = split /\s+/,$_;
						$id = join(":",$temp[0],$temp[1]);
						chomp $_;
						if (/;CDSS=(\d+)$/){
							$score = $1;
							if (/GENE=(.*?);/){
								$anno = $1;
								if (defined $recur_elm{$anno}){							
									s/;CDSS=\d+$//g;
									print OUT $_,";CDSS=",$score+1,";RECUR=",$recur_elm{$anno},"\n";
								}else{
									print OUT $_,"\n";
								}
							}
						}elsif(/NCDS=(\d+)$/){
							$score = $1;
							if (/NCENC=(.*?);/){
								my $line = ""; 
								foreach $anno((split /,/,$1)){
									if (defined $recur_elm{$anno}){
										$line = join("", $line,$recur_elm{$anno},",");
									}
								}
								if (length($line) >0 ){
									s/;NCDS=\d+$//g;	
									$line =~ s/,$//;
									print OUT $_,";NCDS=",$score+1,";RECUR=",$line,"\n";
								}elsif(defined $recur_id_noanno{$id}){
									s/;NCDS=\d+$//g;	
									print OUT $_,";NCDS=",$score+1,";RECUR=",$recur_id_noanno{$id},"\n";
								}else{
									print OUT $_,"\n";
								}
							}else{
								if(defined $recur_id_noanno{$id}){
									s/;NCDS=\d+$//g;	
									print OUT $_,";NCDS=",$score+1,";RECUR=",$recur_id_noanno{$id},"\n";
								}else{
									print OUT $_,"\n";
								}
							}
						}
					}				
				}
			}
			close IN;
			close OUT;	
			unlink ($input);	
		}
	}
}

1;

