#!/usr/bin/perl

use strict;
use IO::Handle;
use Parallel::ForkManager;
use FUNSEQ;

unless(scalar @ARGV == 6){
print <<EOF;
	Missing arguments ... 	
EOF
exit;
}


$| = 1; # Auto Flush
my $infile = $ARGV[0];  
my $maf = $ARGV[1];        # MAF cut-offs
my $genome_mode = $ARGV[2]; # 1: Somatic ; 2: Germline | Personal Genome
my $informat = $ARGV[3];  # User Input Format [BED | VCF]
my $outformat = $ARGV[4]; # Output Format [BED | VCF] 
my $nc_mode=$ARGV[5];    # Whether or not do coding analysis 

#   Check existence of Ouput directory 
my $output_path = "out";
unless (-d $output_path){
	mkdir $output_path;
}

##  Input parameter checking 

die "Error. Please enter proper MAF...\n" unless ($maf =~ /\d+/ && $maf >=0 && $maf <= 1);
die "Error. Please specify correct Genome Mode...\n" unless ($genome_mode == 1 || $genome_mode ==2); 
die "Error. Input format should be bed or vcf...\n" unless ($informat =~ /bed|vcf/i);
die "Error. Output format should be bed or vcf...\n" unless ($outformat =~ /bed|vcf/i);

## Required files
my $file_path = "data";  #File Path
my $tgp_snp = "$file_path/1kg.phase1.snp.bed.gz"; # 1kg SNP File
my $encode_annotation = "$file_path/ENCODE.annotation.gz"; # ENCODE  Annotation
my $cds = "$file_path/gencode7.cds.bed";  # GENCODE v7 CDS
my $bound_motif = "$file_path/ENCODE.tf.bound.union.bed";  # ENCODE TF bound Motif 
my $sensitive = "$file_path/sensitive.nc.bed"; # Sensitive Region
my $promoter = "$file_path/gencode.v7.promoter.bed"; #Promoter Region; upstream +2.5kb
my $enhancer = "$file_path/DRM_transcript_pairs_modify"; #Enhancer Region; 
my $ppi = "$file_path/PPI.hubs.txt";  #Protein protein interaction network hub
my $reg = "$file_path/REG.hubs.txt";  #Regulatory network hub
my $selection = "$file_path/GENE.strong_selection.txt";  #Gene under negative selection 
my $ancestral_file = "$file_path/human_ancestor_GRCh37_e59.fa"; # hg19 ancestral allele
my $coding_interval = "$file_path/gencode.v7.annotation.GRCh37.cds.gtpc.ttpc.interval"; #VAT
my $coding_fasta = "$file_path/gencode.v7.annotation.GRCh37.cds.gtpc.ttpc.fa"; #VAT
my $motif_pfm = "$file_path/motif.PWM";  #Motif PWM files 

## Check required files
die "Error. $tgp_snp not found or empty...\n" unless (-f $tgp_snp && -s $tgp_snp);
die "Error. $encode_annotation not found or empty...\n" unless (-f $encode_annotation && -s $encode_annotation);
die "Error. $cds not found or empty...\n" unless (-f $cds && -s $cds);
die "Error. $bound_motif not found or empty...\n" unless (-f $bound_motif && -s $bound_motif);
die "Error. $sensitive not found or empty...\n" unless (-f $sensitive && -s $sensitive);
die "Error. $promoter not found or empty...\n" unless (-f $promoter && -s $promoter);
die "Error. $enhancer not found or empty...\n" unless (-f $enhancer && -s $enhancer);
die "Error. $ppi not found or empty...\n" unless (-f $ppi && -s $ppi);
die "Error. $reg not found or empty...\n" unless (-f $reg && -s $reg);
die "Error. $selection not found or empty...\n" unless (-f $selection && -s $selection);
die "Error. $ancestral_file not found or empty...\n" unless (-f $ancestral_file && -s $ancestral_file);
die "Error. $motif_pfm not found or empty...\n" unless (-f $motif_pfm && -s $motif_pfm);
if ($nc_mode==0){
	die "Error. $coding_interval not found or empty...\n" unless (-f $coding_interval && -s $coding_interval);
	die "Error. $coding_fasta not found or empty...\n" unless (-f $coding_fasta && -s $coding_fasta);
}

## If number of input files larger than 1, do parallele running. 
my @input = split /,/,$infile;
if (scalar @input ==1){
	&funseq($input[0]);
}else{
	my $num_per_run = 5;	# number of genome per run. You can change this according to your system. 
	my $MAX_PROCESSES = 10;   
	my $pm = new Parallel::ForkManager($MAX_PROCESSES);
	my $data;
	my $i;
	my $j;
	my $num = int(scalar @input / $num_per_run);

        # parallel : maximum $num_per_run genomes per group. 
        for $j (1 .. $num){
                for $i (($j-1)*$num_per_run  .. $j*$num_per_run - 1){
                		sleep(3);
                        my $pid = $pm -> start and next;
                        &funseq($input[$i]);
                        $SIG{INT} = sub {kill 9, $pid;};
                                $pm -> finish;
                }
                $pm -> wait_all_children;
        }

        if ((scalar @input - $num*$num_per_run )>0){
                for $i ($num*$num_per_run  .. $#input){
                		sleep(3);
                        my $pid = $pm -> start and next;
                        &funseq($input[$i]);
                        $SIG{INT} = sub {kill 9, $pid;};
                                $pm -> finish;
                }
                $pm -> wait_all_children;
        }

	
	FUNSEQ::recur(\@input,$output_path,$outformat);
}







#   Pipeline 
sub funseq{
	my ($infile) = @_;
	my @temp = split /\//,$infile;
	my $tag = (split /\./,$temp[$#temp])[0];
	# Record Log and Error.
	my $log = "$output_path/$tag.log";
	my $err = "$output_path/$tag.err";
	open(LOG,">$log")||die;
	open(ERR,">$err")||die;
	STDERR->fdopen( \*ERR,  'w' ) or die $!;
	
	# Intermedia files. 
	my $out_snv_filter= "$output_path/$tag.out.snv.filter";
	my $out_nc = "$output_path/$tag.out.nc";
	my $out_annotated_extend = "$output_path/$tag.out.annotated.extend";
	my $out_cds = "$output_path/$tag.out.cds";
	my $out_vat = "$output_path/$tag.out.vat";

	# Output File
	my $funseq_out = "$output_path/$tag.FunSEQ.$outformat";
	my $status;
 	
 	# check for input file format 
 	die "Error. Input file not found or empty...\n" unless (-f $infile && -s $infile);
    	print "... Input format check : $informat ...\n";
    	$status = FUNSEQ::format_check($infile,$informat);
	if ($status ==0){
		print "... Format ok ...\n";
	}else{
		exit;
	}
	
	# Pipeline starts
	print "... $tag Start (log file : $output_path/$tag.log ; error file : $output_path/$tag.err) ...\n";
	print LOG "FunSEQ -f $tag -maf $maf -m $genome_mode -inf $informat -outf $outformat\n","-"x30,"\n";
	print LOG `date`,"Start running pipeline ...\n";
	

	# * Step1: 1000 genomes SNV filtering 
	print  LOG `date`,"Start filtering SNVs with minor allele frequency = $maf in 1000 genomes phase1 ...\n";
	
	my $data = new FUNSEQ;
	$data -> snv_filter($infile,$informat,$tgp_snp,$maf,$out_snv_filter);
	if (-z $out_snv_filter){
		print "Error: Filter against 1000 genomes SNV data, no variants left ...\n";
		`rm $output_path/$tag.out*`;
		exit;
	}	
	
	print LOG `date`,"Filtering Done\n";
	
	
	print "...... 30% Complete ......";
	print "\r";
		
	
	# * Step2: Get Annotated (Coding & Non-coding) SNVs
	print LOG `date`,"Separate into non-coding annotation & Coding  ...\n";
	
	`intersectBed -u -a $out_snv_filter -b $cds > $out_cds`; # Obtain coding variants (based on Gencode 7)
	if (-s $out_cds){
            `intersectBed -v -a $out_snv_filter -b $out_cds > $out_nc`;
    }else{
            `cp $out_snv_filter $out_nc`;
    }

	if (-s $out_nc){
		`intersectBed -a $out_nc -b $encode_annotation -wo > $out_annotated_extend`;
	}

	print "...... 50% Complete ......";
	print "\r";


	# * Step 3: Non-coding 
	print LOG `date`,"Start Non-coding SNVs analysis ...\n";
	print LOG `date`,"1. Motif Breaking ...\n";
	
	if (-s $out_nc){
		$data -> sensitive($out_nc,$sensitive,$informat);
		$data -> motif_break($out_nc, $informat, $ancestral_file, $bound_motif, $genome_mode, $motif_pfm);
		$data -> gene_link($out_nc,$informat,$promoter,$enhancer);
	}

	print "...... 70% Complete ......";
	print "\r";


	# * Step 4: Coding 
	print LOG `date`,"Start Coding SNVs analysis ...\n";
	
	if (-z $out_cds){
	}else{
		$status = FUNSEQ::coding($out_cds,$informat,$coding_interval,$coding_fasta,$ppi,$reg,$selection,$out_vat,$output_path,$nc_mode);
		if ($status == 1){
			print "Error: VAT problem ... \n";
			`rm $output_path/$tag.out*`;
			exit;
		}
	}
	
	print "...... 90% Complete ......";
	print "\r";


	# * Step 5: Scoring SNVs 
	$data -> intergrate($informat,$outformat,$tag,$out_nc,$out_annotated_extend,$out_vat,$funseq_out);
	print LOG `date`,"FunSEQ Done\n";
	print "... $tag Done ...\n";
	
	
	# delete all intermedia results 
	my $choice = 1; # 1 Yes; Other No
	if ($choice == 1){
		 `rm $output_path/$tag.out*`;
	}
	close LOG;
	close ERR;

}   

          
                




